#!/usr/bin/env python
# encoding: utf-8
"""
polish_pipeline.py

Created by Rod Docking on 2011-02-08.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import subprocess
import shutil
from optparse import OptionParser, OptionGroup
from time import sleep
from glob import glob
import logging
try:
    from ruffus import *
except ImportError, E:
    print "Failed to import ruffus: ", E
    raise E

from lib.cluster_job import Cluster_job
from lib.check_prerequisites import check_prerequisites

PYTHON_EXECUTABLE  = sys.executable
LAUNDRY_SCRIPT     = os.path.dirname(sys.argv[0]) + '/lib/fasta_laundry.py'
SCRUB_SCRIPT       = os.path.dirname(sys.argv[0]) + '/anchor_snp_indel_corrector.py'
ANCHOR_SCRIPT      = os.path.dirname(sys.argv[0]) + '/anchor_scaffold_filler.py'
DIVERT_SCRIPT      = os.path.dirname(sys.argv[0]) + '/lib/divert_static_alignments.py' 
INSERT_SIZE_SCRIPT = os.path.dirname(sys.argv[0]) + '/lib/infer_isize_and_read_length.py'
REHEADER_SCRIPT    = os.path.dirname(sys.argv[0]) + '/lib/restore_fasta_headers.py'

# TO-DO: Decide where to set the Picard directory - as an environment variable?
PICARD_DIR = os.environ['PICARD_DIR']
MPILEUP_REGION_SIZE = 50000000

##################################################
#
#   Option parsing and global names
#
##################################################

#
# Parse command-line options
#

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] original_ref.fa OUTPUT_ROOT READS_LIST"
    parser = OptionParser(usage)
    
    group_one = OptionGroup(parser,"Tasks")
    group_one.add_option("-S", "--scrub",
        action="store_true", dest="scrub",
        help="Scrub contigs to correct poorly-supported base calls and indels")
    group_one.add_option("-A", "--anchor",
        action="store_true", dest="anchor",
        help="Run the anchor local assembler to fill in scaffold gaps")
    parser.add_option_group(group_one)

    group_two = OptionGroup(parser,"Run Modes")
    group_two.add_option("-I", "--iterate",
        action="store_true", dest="iterate",
        help="Run in iterative mode")
    group_two.add_option("-F", "--finalize",
        action="store_true", dest="finalize",
        help="Produce final output files")
    parser.add_option_group(group_two)

    parser.add_option("-c", "--cluster",
        action="store_true", dest="cluster",
        help="Run jobs on an SGE cluster instead of locally")
    parser.add_option("-b", "--bam",
        action="store_true", dest="bam",
        help="Reads are stored in BAM files")
    parser.add_option("-r", "--recon",
        action="store_true", dest="recon",
        help="Show jobs that need to be run, but don't actually run them")
    parser.add_option("-d", "--dirty",
        action="store_true", dest="dirty",
        help="Leave all intermediate files in place")
    
    (opt, args) = parser.parse_args()
    
    if not (len(args) == 3):
        parser.error("incorrect number of arguments")

    if opt.iterate and opt.finalize:
        parser.error("Iterate and Finalize modes are mutually exclusive")
    
    if opt.finalize and (opt.scrub or opt.anchor):
        parser.error("No assembly-cleaning tasks can be performed for Finalize mode")
    
    return opt, args

# Parse options
opt, args = parse_options()

# Check pre-requisites after showing usage info
check_prerequisites()

original_ref, output_root, reads_list = args

#
# Set up names for output files based on original reference file and input reads files
#

if opt.iterate or opt.finalize:
    
    assembly_files = glob('{0}-*.fa'.format(output_root))
    
    assembly_numbers = []
    for assembly in assembly_files:
        
        f1 = assembly.split('-')
        f2 = f1[-1].split('.')
        try:
            assembly_numbers.append(int(f2[0]))
        except ValueError, e:                       # Skip over files without integers
            continue
    
    try:
        max_assembly = max(assembly_numbers)
        input_suffix = str(max_assembly)
        scrub_suffix = str(max_assembly + 1)
        anchor_suffix = str(max_assembly + 2)
    except ValueError:
        print "No previous assemblies found"
        input_suffix, scrub_suffix, anchor_suffix = "10", "11", "12"
    
    print "Assembly numbers: ", assembly_numbers
    print "Input: ", input_suffix, "Scrub: ", scrub_suffix, "Anchor: ", anchor_suffix

# For the first iteration (i.e., coming from the raw ABySS assembly)
else:
    input_suffix, scrub_suffix, anchor_suffix = "10", "11", "12"
    
# Names for reference files
cleaned_ref = output_root + "-{0}.fa".format(input_suffix)

scrubbed_ref = output_root + "-{0}.fa".format(scrub_suffix)
scrubbed_ref_touchfile = scrubbed_ref + ".complete"

anchored_ref = output_root + "-{0}.fa".format(anchor_suffix)
anchored_ref_touchfile = anchored_ref + ".complete"

# For the last round, set up special names    
if opt.finalize:
    
    # This is a bit obscure - think of a better way to do it
    final_ref = output_root + "-anchor.fa"  
    shutil.copy(cleaned_ref, final_ref)
    cleaned_ref = output_root + "-anchor.fa"
    input_suffix = "anchor"
    #anchor_suffix = "anchor"

final_ref_raw = output_root + "-anchor.raw.fa"
final_merged_bam = output_root + "-anchor.bam"

cleaned_ref_faidx = cleaned_ref + ".fai"
cleaned_ref_bwa_index = cleaned_ref + ".rsa"

# Go through the list of read files
read_files = []
for line in open(reads_list, 'r'):
    c = line.split()
    read_files.extend(c)

# For now, we're assuming that all read files are paired and in separate files
# BAM files should only be listed once in the input file
if not opt.bam:
    assert(len(read_files) % 2 == 0)

# mpileup output files
raw_bcf = output_root + "-{0}.raw.bcf".format(input_suffix)
filtered_vcf = output_root + "-{0}.flt.vcf".format(input_suffix)

# Names for alignment files and 'touch' files
read_pairs, sai_files, sai_pairs, bam_files, sorted_bam_files, bai_files = [], [], [], [], [], []
aln_touch_pairs, sampe_touch_files, picard_touch_files, bam_insert_files, i = [], [], [], [], 1
static_bam_files, realign_bam_files = [], []

# Intermediate files for splitting mpileup jobs
region_root = output_root + "-{0}".format(input_suffix)

# This list of bam files is used by mpileup
bam_list = reads_list + ".bamlist"
bam_list_handle = open(bam_list, 'w')

# Infer pairs for read files and sai files and generate intermediate file names
for i in range(0, len(read_files), 2):
    
    # If we're iterating, the reads to realign are in files called *.realign.bam
    if opt.iterate and not (input_suffix == '10'):
        previous_static_bam = output_root + '-{0}_{1}.realign.bam'.format( input_suffix, i)
        read_pairs.append( (previous_static_bam, previous_static_bam) )
    
    elif opt.bam:
        
        # BAM files will be listed singly in the input file
        read_pairs.append( (read_files[i], read_files[i]) )
        
        # Need to look ahead in the list here
        if not i == len(read_files)-1:
            read_pairs.append( (read_files[i+1], read_files[i+1]) )
    else:
        read_pairs.append( (read_files[i], read_files[i+1]) )

for i in range(len(read_pairs)):
    
    # Extend the list of all sai file names, and append the pair to the list of pairs
    pathed_roots = [os.path.join(os.getcwd(), os.path.basename(r)) for r in (read_pairs[i][0], read_pairs[i][1]) ]
    sai_pair = [pathed_roots[0] + ".1.sai", pathed_roots[1] + ".2.sai"]
    sai_files.extend(sai_pair)
    sai_pairs.append(sai_pair)
    
    # touch files to indicate alignment jobs are complete
    touch_pair = [pathed_roots[0] + ".1.sai.complete", pathed_roots[1] + ".2.sai.complete"]
    aln_touch_pairs.append( touch_pair )
    
    # bam files and touch files for bwa sampe - these use the output root..
    bam_file = output_root + '-{0}_{1}.bam'.format( input_suffix, i )
    static_bam_file = output_root + '-{0}_{1}.static.bam'.format( anchor_suffix, i )
    realign_bam_file = output_root + '-{0}_{1}.realign.bam'.format( anchor_suffix, i )
    
    bam_files.append(bam_file)
    static_bam_files.append(static_bam_file)
    realign_bam_files.append(realign_bam_file)
    sampe_touch_files.append( output_root + '-{0}_{1}.bam.complete'.format( input_suffix, i) )

    # Picard intermediate files
    sorted_bam_file = output_root + '-{0}_{1}.sorted.bam'.format( input_suffix, i)
    sorted_bam_files.append(sorted_bam_file)
    
    insert_stat_file = output_root + '-{0}_{1}.sorted.insert.txt'.format( input_suffix, i)
    bam_insert_files.append(insert_stat_file)

    bai_file = output_root + '-{0}_{1}.sorted.bai'.format( input_suffix, i )
    bai_files.append(bai_file)
    picard_touch_files.append(bai_file + ".complete")
    
    bam_list_handle.write("{0}\n".format(sorted_bam_file))
    
bam_list_handle.close()

# edits_list is a file that lists all the contigs that have been edited this round
edits_list = output_root + "-{0}.edited_contigs.txt".format(anchor_suffix) 

# Set up logfile
logger = logging.getLogger("anchor_pipeline")
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler("{0}.log".format(reads_list), 'w')
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

# Redirect standard error and standard out for all subprocess jobs
std_error_log = open('{0}.err'.format(reads_list),'w')
std_out_log = open('{0}.out'.format(reads_list), 'w')
summary_file = output_root + "-{0}.results".format(anchor_suffix)

############################################################
#
#   Pipeline functions handled by Ruffus
#
############################################################

#
#   Clean the starting fasta file (this should show up as up-to-date if we iterate)
#

@files(original_ref, cleaned_ref)
def clean_fasta(input_file, output_file):
    """Strip comments and add line breaks to a fasta file"""
    
    logger.debug("start clean_fasta")
    
    args = [PYTHON_EXECUTABLE, LAUNDRY_SCRIPT, '-N', '-l', '50', input_file, output_file]
    logger.info(" ".join(args))
    
    proc = subprocess.Popen(args, shell=False, cwd=os.getcwd(), 
                            stderr=std_error_log, stdout=std_out_log)
    proc.wait()
    
    logger.debug("finished clean_fasta")
    return

#
#   Index the cleaned fasta file
#

@follows(clean_fasta)
@files(cleaned_ref, cleaned_ref_faidx)
def samtools_faidx(reference, index_file):
    """Run samtools faidx"""
    
    logger.debug("start samtools_faidx")
    
    # Run samtools faidx to index reference file for sequence retrieval 
    args = ['samtools', 'faidx', reference]
    logger.info(" ".join(args))
     
    proc = subprocess.Popen(args, shell=False, cwd=os.getcwd(), 
                            stderr=std_error_log, stdout=std_out_log)
    proc.wait()
    
    logger.debug("finished samtools_faidx")
    return

#
#   Index the cleaned fasta file with bwa index
#

@follows(samtools_faidx)
@files(cleaned_ref, cleaned_ref_bwa_index)
def bwa_index(input_file, output_file):
    """Run bwa index"""
    
    logger.debug("start bwa_index")
    
    # Use different algorithm depending on file-size:
    ref_size = os.path.getsize(cleaned_ref)
    if ref_size < 1000000000:
        algo = 'is'
        logger.info('Using bwa index is')
    else:
        algo = 'bwtsw'
        logger.info('Using bwa index bwtsw')
    
    # Run bwa index to index reference file for BWA alignments
    args = ['bwa', 'index', '-a', algo, input_file]
    logger.info(" ".join(args))
    
    proc = subprocess.Popen(args, shell=False, cwd=os.getcwd(), 
                            stderr=std_error_log, stdout=std_out_log)
    proc.wait()

    logger.debug("finished bwa_index")
    return

#
#   Run bwa aln for both reads files in parallel
#

parameters = [ [x,y,z] for x,y,z in zip(read_pairs, sai_pairs, aln_touch_pairs)]
@follows(bwa_index)
@files(parameters)
def bwa_aln(read_pair, sai_pair, touch_pair):
    """Run bwa aln"""
    
    logger.debug("start bwa_aln")
    print read_pair
    
    for i in [0,1]:
        
        # Remove any old completion files
        if os.path.exists(touch_pair[i]):
            os.unlink(touch_pair[i])
                
        if opt.cluster:
            
            # If we're using BAM files (if initial input was BAM, or we're not on the first iteration)
            if opt.bam or (opt.iterate and not (input_suffix == '10')):
                cmd = "bwa aln -b{0} -t12 {1} {2} > {3}".format(i+1, cleaned_ref, read_pair[i], sai_pair[i])
            
            # If we're using fastq files    
            else:
                cmd = "bwa aln -t12 {0} {1} > {2}".format(cleaned_ref, read_pair[i], sai_pair[i])
            
            touch_cmd = "touch {0}".format(touch_pair[i])
            logger.info(cmd)
            logger.info(touch_cmd)
            
            jobname = "bwa_aln_{0}_{1}.sh".format(os.path.basename(read_pair[i]), i+1)
            C = Cluster_job(jobname = jobname, pe = 'ncpus', cpus = 12)
            C.append(cmd)
            C.append(touch_cmd)
            C.submit()
        
        else:
            
            args = []

            # If we're using BAM files (if initial input was BAM, or we're not on the first iteration)
            if opt.bam or (opt.iterate and not (input_suffix == '10')):
                args = ['bwa', 'aln', '-b{0}'.format(i+1), '-t2', cleaned_ref, read_pair[i]]

            # If we're using fastq files
            else:
                args = ['bwa', 'aln', '-t2', cleaned_ref, read_pair[i]]
            
            logger.info(" ".join(args))
            proc = subprocess.Popen(args, shell=False, cwd=os.getcwd(), 
                                    stdout=open(sai_pair[i], 'wb'), stderr=std_error_log)
            proc.wait()
                                    
            fh = open(touch_pair[i], 'w')
            fh.close()
    
    logger.debug("finish bwa_aln")
    return

#
#   Check for all bwa aln jobs to be finished
#

parameters = [ [x,y] for x,y in zip(read_files, aln_touch_pairs)]
@follows(bwa_aln)
@files(parameters)
def bwa_aln_wait(read_file, touch_pair):
    """Wait until all the .sai files are generated"""
    
    logger.debug("start bwa_aln_wait")
    
    # Wait for cluster jobs to be complete
    print "Waiting for align jobs"
    sleep(60)
    
    while True:
        
        # Wait for each file to complete
        if not os.path.exists(touch_pair[0]):
            std_error_log.write("{0} does not exist yet - waiting...\n".format(touch_pair[0]))
            sleep(60)
        elif not os.path.exists(touch_pair[1]):
            std_error_log.write("{0} does not exist yet - waiting...\n".format(touch_pair[1]))
            sleep(60)
        
        # Once they're done, break
        else:
            break
    
    logger.debug("finish bwa_aln_wait")
    return

#
#   Run bwa sampe to generate a sorted BAM file
#

parameters = [ [w,x,y,z] for w,x,y,z in zip(sai_pairs, bam_files, read_pairs, sampe_touch_files)]
@follows(bwa_aln_wait)
@files(parameters)
def bwa_sampe(sai_pair, bam_file, read_pair, touch_file):
    """Run bwa sampe"""
    
    logger.debug("start bwa_sampe")
    
    # Remove any old completion files
    if os.path.exists(touch_file):
        os.unlink(touch_file)

    # Check that sai files are not empty
    for sai in sai_pair:
        assert(os.path.getsize(sai) > 1000)
            
    if opt.cluster:
        
        RG_line = "\'@RG\tID:1\tSM:1\'"
        
        sampe_cmd = "bwa sampe -s -r {0} {1} {2} {3} {4} {5} | \\".format(
            RG_line, cleaned_ref, sai_pair[0], sai_pair[1], read_pair[0], read_pair[1])
        view_cmd = "\tsamtools view -bS - > {0}".format(bam_file)
        touch_cmd = "touch {0}".format(touch_file)
        
        jobname = "bwa_sampe_{0}.sh".format(os.path.basename(bam_file))
        
        logger.info(sampe_cmd)
        logger.info(view_cmd)
                
        C = Cluster_job(jobname = jobname, mem_token = "10G")
        C.append(sampe_cmd)
        C.append(view_cmd)
        C.append(touch_cmd)
        C.submit()
    
    else:
        
        RG_line = "@RG\tID:1\tSM:1"
        
        args1 = ['bwa', 'sampe', '-s', '-r', RG_line, cleaned_ref, sai_pair[0], sai_pair[1], read_pair[0], read_pair[1]]
        proc1 = subprocess.Popen(args1, shell=False, cwd=os.getcwd(), 
                                stdout=subprocess.PIPE, stderr=std_error_log)
        logger.info(" ".join(args1))

        args2 = ['samtools', 'view', '-bS', '-']
        proc2 = subprocess.Popen(args2, shell=False, cwd=os.getcwd(),
                                stdin=proc1.stdout, stdout=open(bam_file, 'wb'), 
                                stderr=std_error_log)
        logger.info(" ".join(args2))
        
        proc2.wait()
                
        fh = open(touch_file, 'w')
        fh.close()
    
    logger.debug("finish bwa_sampe")
    return

# 
#   Wait for bwa sampe jobs to finish
# 

parameters = [ [x[0],y] for x,y in zip(sai_pairs, sampe_touch_files)]
@follows(bwa_sampe)
@files(parameters)
def bwa_sampe_wait(sai_file, touch_file):
    """Wait until all the BAM files are generated"""
    
    logger.debug("start bwa_sampe_wait")
    print "Waiting for sampe job ", touch_file
    sleep(60)
    
    while True:
        
        if not os.path.exists(touch_file):
            std_error_log.write("{0} does not exist yet - waiting...\n".format(touch_file))
            sleep(60)
        else:
            break
    
    logger.debug("finish bwa_sampe_wait")
    return

#
#   Run Picard FixMateInformation to sort, clean mates, and index the BAM file
#

parameters = [ [x,y,z] for x,y,z in zip(bam_files, sorted_bam_files, picard_touch_files) ]
@follows(bwa_sampe_wait)
@files(parameters)
def picard_fixmate(bam_file, sorted_bam_file, picard_touch_file):
    """ Run Picard FixMateInformation.jar to fix mate information, sort, and index the BAM file """

    logger.debug("start picard_fixmate")
    
    # Remove any old completion files
    if os.path.exists(picard_touch_file):
        os.unlink(picard_touch_file)
            
    picard_cmd = """java -jar -Xmx40G -XX:MaxPermSize=20g {0}/FixMateInformation.jar INPUT={1} OUTPUT={2} SO=coordinate CREATE_INDEX=true VALIDATION_STRINGENCY=SILENT""".format(
                  PICARD_DIR, bam_file, sorted_bam_file)
    
    flagstat_file = sorted_bam_file[:-4] + ".flagstat"               
    flagstat_cmd = "samtools flagstat {0} > {1}".format(sorted_bam_file, flagstat_file)
    
    logger.info(picard_cmd)
    logger.info(flagstat_cmd)
    
    if opt.cluster:
        
        jobname = "picard_{0}.sh".format(bam_file)
        C = Cluster_job(jobname = jobname, mem_token = "40G", tmp_token = "60G")
        C.append(picard_cmd)
        C.append(flagstat_cmd)
        C.append("touch {0}".format(picard_touch_file))
        C.submit()
        
    else:
        
        picard_args = picard_cmd.split()
        
        proc = subprocess.Popen(picard_args, shell=False, cwd=os.getcwd(), stderr=std_error_log)
        proc.wait()
        proc2 = subprocess.Popen(['samtools', 'flagstat', sorted_bam_file], shell=False, cwd=os.getcwd(), 
                                stdout=open(flagstat_file, 'w'), stderr=std_error_log)
        proc2.wait()
        
        open(picard_touch_file, 'w')

    logger.debug("finish picard_fixmate")
    return

#
#   Wait for all Picard jobs to finish
#

parameters = [ [x,y] for x,y in zip(bam_files, picard_touch_files)]
@follows(picard_fixmate)
@files(parameters)
def picard_fixmate_wait(bam_file, picard_touch_file):
    """ Wait for all picard jobs to finish"""
    
    logger.debug("start picard_fixmate_wait")
        
    # Check for cluster jobs to be complete, and tidy up submission files
    print "Waiting for Picard job ", picard_touch_file
    sleep(60)
    
    while True:
        
        if not os.path.exists(picard_touch_file):
            std_error_log.write("{0} does not exist yet - waiting...\n".format(picard_touch_file))
            sleep(60)
        else:
            break
    
    logger.debug("start picard_fixmate_wait")
    return
    
#
#   Generate regions files to split up mpileup jobs
#

@follows(picard_fixmate_wait)
@split(cleaned_ref_faidx, "*.region_*.txt")
def generate_regions_files(cleaned_ref_faidx, regions):
    """ Generate a list of regions to split mpileup jobs """
    
    logger.debug("start generate_regions_files")
        
    region_counter = 0
    sequence_sum = 0
    region_file = region_root + ".region_{0}.txt".format(region_counter)
    region_handle = open(region_file, 'w')
    
    with open(cleaned_ref_faidx, 'r') as fai_handle:
        
        for line in fai_handle:
            cols = line.split()
            c_id, c_len = cols[0], int(cols[1])
            
            # Open a new file if we need to
            if not region_handle:
                region_file = region_root + ".region_{0}.txt".format(region_counter)
                region_handle = open(region_file, 'w')
            
            # NOTE: this requires at least samtools 933:176 (treating mpileup -l like a BED file)
            # i.e., at least samtools 0.1.15
            region_handle.write("{0}\t{1}\t{2}\n".format(c_id, '1', c_len))
            
            # Then check if we need to start a new file
            sequence_sum += c_len
            if sequence_sum >= MPILEUP_REGION_SIZE:
                region_handle.close()
                region_counter += 1
                sequence_sum = 0
                region_handle = None
    
    logger.info("Generated {0} region files, last is {1}".format(region_counter, region_file))
    logger.debug("finish generate_regions_files")
    return

#
#   Run samtools mpileup to generate SNP/indel calls
#

@follows(generate_regions_files)
@transform(generate_regions_files, suffix(".txt"), ".part.bcf", bam_list)
def samtools_mpileup(region_file, partial_bcf_file, bam_list):
    """Run samtools mpileup to generate raw SNP calls"""
    
    logger.debug("start samtools_mpileup")
            
    partial_bcf_touch_file = partial_bcf_file + ".complete"
    if os.path.exists(partial_bcf_file):
        os.unlink(partial_bcf_file)
    
    # Example
    # samtools mpileup -ugf ref.fa aln1.bam aln2.bam | bcftools view -bvcg - > var.raw.bcf
    
    if opt.cluster:
        
        mpileup_cmd = 'samtools mpileup -C50 -d10000 -Augf {0} -b {1} -l {2} |\\'.format(
            cleaned_ref, bam_list, region_file)
        view_cmd = '  bcftools view -bvcg - > {0}'.format(partial_bcf_file)
        touch_cmd = "touch {0}".format(partial_bcf_touch_file)
        
        jobname = "samtools_mpileup_{0}.sh".format(os.path.basename(region_file))
        
        C = Cluster_job(jobname = jobname, mem_token = "10G")

        logger.info(mpileup_cmd)
        logger.info(view_cmd)
        logger.info(touch_cmd)

        C.append(mpileup_cmd)
        C.append(view_cmd)
        C.append(touch_cmd)
        C.submit()
    
    else:
        
        mpileup_args = ['samtools', 'mpileup', '-C50', '-d10000', '-Augf', cleaned_ref,
                        '-b', bam_list, '-l', region_file]
        logger.info(" ".join(mpileup_args))
        
        proc1 = subprocess.Popen(mpileup_args, shell=False, cwd=os.getcwd(), 
                                stdout=subprocess.PIPE, stderr=std_error_log)
        
        bcftools_args = ['bcftools', 'view', '-bvcg', '-']
        logger.info(" ".join(bcftools_args))
        
        proc2 = subprocess.Popen(bcftools_args,
            shell=False, cwd=os.getcwd(), stdin=proc1.stdout,
            stdout=open(partial_bcf_file, 'wb'), stderr=std_error_log
            )
        proc2.wait()
        
        fh = open(partial_bcf_touch_file, 'w')
        fh.close()
    
    logger.debug("finish samtools_mpileup")
    return

#
#   Wait for all mpileup jobs to complete
#

@follows(samtools_mpileup)
@transform(generate_regions_files, suffix(".txt"), ".part.bcf.complete")
def samtools_mpileup_wait(region_file, partial_bcf_touch_file):
    """ Wait for all samtools mpileup jobs to complete """
    
    logger.debug("start samtools_mpileup_wait")
        
    sleep(60)
    print "Waiting for mpileup job ", partial_bcf_touch_file
    while True:
        
        if not os.path.exists(partial_bcf_touch_file):
            std_error_log.write("{0} does not exist yet - waiting...\n".format(partial_bcf_touch_file))
            sleep(60)
        else:
            break
    
    logger.debug("finish samtools_mpileup")
    return

#
#   Run bcftools cat to merge all the .tmp.bcf files
#

@follows(samtools_mpileup_wait)
@merge(samtools_mpileup_wait, raw_bcf)
def bcftools_cat(partial_bcf_files, raw_bcf):
    """ Concatenate all the partial raw BCF files """
    
    logger.debug("start bcftools_cat")
    
    # Count the BCF files, and assert that they're not empty
    bcf_files = glob('*.part.bcf')
    logger.info("Found {0} partial BCF files".format(len(bcf_files)))
        
    # Regenerate the bcf file name from the touch-file name
    partial_bcf_files = [x.rstrip('.complete') for x in partial_bcf_files]

    args = ['bcftools', 'cat'] + partial_bcf_files
    logger.info(" ".join(args))
    
    proc1 = subprocess.Popen(args, shell=False, cwd=os.getcwd(),
            stdout=open(raw_bcf, 'wb'), stderr=std_error_log)
    proc1.wait()
    
    logger.debug("finish bcftools_cat")
    return

#
#   Run bcftools view to filter SNP calls
#

@follows(bcftools_cat)
@files(raw_bcf, filtered_vcf)
def bcftools_view(raw_bcf, filtered_vcf):
    """Run bcftools view | vcfutils.pl varFilter to filter SNP calls"""
    
    logger.debug("start bcftools_view")
        
    # bcftools view var.raw.bcf | vcfutils.pl varFilter -D100 > var.flt.vcf
    
    bcftools_args =  ['bcftools', 'view', raw_bcf]
    logger.info(" ".join(bcftools_args))
    
    proc1 = subprocess.Popen(bcftools_args, shell=False, cwd=os.getcwd(), 
            stdout=subprocess.PIPE, stderr=std_error_log)
    
    vcfutils_args = ['vcfutils.pl', 'varFilter', '-D10000', '-p', '-2', '0']
    logger.info(" ".join(vcfutils_args))
    
    proc2 = subprocess.Popen(vcfutils_args,
        shell=False, cwd=os.getcwd(), stdin=proc1.stdout,
        stdout=open(filtered_vcf, 'w'), stderr=std_error_log)
    
    proc2.wait()
    
    logger.debug("finish bcftools_view")
    return

#
#   Run the Anchor SNP/indel corrector
#

@follows(bcftools_view)
@files(filtered_vcf, scrubbed_ref, cleaned_ref, edits_list)
def scrub_assembly(filtered_vcf, scrubbed_ref, cleaned_ref, edits_list):
    """ Run the Anchor SNP/Indel correction tool """
    
    if opt.scrub:
        logger.debug("start scrub_assembly")
    
        scrub_args = [PYTHON_EXECUTABLE, SCRUB_SCRIPT, '-oi',
                        filtered_vcf, cleaned_ref, scrubbed_ref, edits_list]
        logger.info(" ".join(scrub_args))
                        
        proc = subprocess.Popen(scrub_args, shell=False, cwd=os.getcwd(), 
                                stderr=std_error_log, stdout=std_out_log)
        proc.wait()
    
        logger.debug("finish scrub_assembly")

    # This should change!
    else:
        logger.debug("not scrubbing contigs - Scrub not requested")
        shutil.copy(cleaned_ref, scrubbed_ref)

    return

#
#   Calculate insert size and read length for each BAM file
#

parameters = [ [x,y] for x,y in zip(sorted_bam_files, bam_insert_files)]
@follows(scrub_assembly)
@files(parameters)
def infer_insert_size(sorted_bam_file, bam_insert_file):
    """ For each BAM file, check the read length and average insert size """
    
    logger.debug("start infer_insert_size")
    
    if opt.anchor:
        
        args = [PYTHON_EXECUTABLE, INSERT_SIZE_SCRIPT, 
                sorted_bam_file, bam_insert_file]
        logger.info(" ".join(args))
                
        proc = subprocess.Popen(args, shell=False, cwd=os.getcwd(), 
                                stderr=std_error_log, stdout=std_out_log)
        proc.wait()

    else:
        logger.debug("not estimating insert size - Anchor not requested")
        fh = open(bam_insert_file,'w')
        fh.close()

    logger.debug("finish infer_insert_size")
    return

#
#   Run the Anchor assembler to close gaps
#

@follows(infer_insert_size)
@files(scrubbed_ref, anchored_ref, anchored_ref_touchfile, bam_list, edits_list, bam_insert_files[0])
def anchor_gap_close(scrubbed_ref, anchored_ref, anchored_ref_touchfile, bam_list, edits_list, bam_insert_file):
    """ Run the anchor gap-closing module """

    logger.debug("start anchor_gap_close")

    if opt.anchor:

        # Read the bam insert file to set anchor parameters
        read_length, isize = open(bam_insert_file).readlines()[0].rstrip().split()

        # Set overlap at 60% of read length
        max_overlap = str(int(int(read_length) * 0.6))

        logger.info("Anchor inferred parameters: -r {0} -i {1} -o {2}".format(
                    read_length, isize, max_overlap))

        # TO-DO - can we split this off as cluster jobs?
        # TO-DO - need to handle multiple fragment-size libraries here - will need to update Anchor script
        #       - currently we're just using the FIRST bam file

        args = [PYTHON_EXECUTABLE, ANCHOR_SCRIPT, '-r', read_length, '-i', isize, 
                '-c', "2", '-o', max_overlap, scrubbed_ref, bam_list, 
                anchored_ref, edits_list]
        logger.info(" ".join(args))
        proc = subprocess.Popen(args,shell=False, cwd=os.getcwd(), 
                                stderr=std_error_log, stdout=std_out_log)
        proc.wait()

    else:
        logger.debug("not filling scaffolds - Anchor not requested")
        shutil.copy(scrubbed_ref, anchored_ref)

    logger.debug("finish anchor_gap_close")
    return

#
#   Split query-sorted alignments into 'static' and 'realign' files
#

parameters = [ [x,y,z, edits_list] for x,y,z in zip(sorted_bam_files, static_bam_files, realign_bam_files)]
@follows(anchor_gap_close)
@files(sorted_bam_file, static_bam_file, realign_bam_file)
def divert_alignments(sorted_bam_file, static_bam_file, realign_bam_file):
    """ Divert static alignments from those that need to be re-aligned """
    
    logger.debug("start divert alignments")

    if opt.iterate:
        
        args = [PYTHON_EXECUTABLE, DIVERT_SCRIPT, edits_list, sorted_bam_file, 
                static_bam_file, realign_bam_file]
        logger.info(" ".join(args))
    
        proc = subprocess.Popen(args, shell=False, cwd=os.getcwd(),  
                                stderr=std_error_log, stdout=std_out_log)
        proc.wait()
    
    else:
        
        logger.debug("Not producing separated alignment files - not running in iterative mode")
        fh = open(realign_bam_file, 'w')
        fh.close()
        
    logger.debug("finish divert alignments")
    return

#
#   Summarize results
#

@follows(divert_alignments)
@files(realign_bam_file, summary_file, input_suffix, anchored_ref)
def summarize_results(realign_bam_file, summary_file, input_suffix, anchored_ref):
    """ Summarize the results for this round """

    logger.debug("start summarize_results")

    std_out_log.close()
    summary_handle = open(summary_file, 'w')
    summary_handle.write("||SNV||Indel||Num Gaps||Bases in Gaps||Merge||Extend||Fail||Mapped (%)||Paired (%)|| Assembly\n")
    result_list = []

    # This goes through the stdout output of the scrubbing, anchor, and flagstat commands
    with open('{0}.out'.format(reads_list), 'r') as result_handle:
        for line in result_handle:
            if "SCRUB" in line:
                c = line.split()
                result_list.extend([c[1], c[2]])
            elif "ANCHOR_1" in line:
                c = line.split()
                result_list.extend([c[1], c[2]])
            elif "ANCHOR_2" in line:
                c = line.split()
                result_list.extend([c[1], c[2], c[3]])
            
    # Find the flagstat files - note that we're just checking the first one here!
    flagstat_file = glob('*{0}_0.sorted.flagstat'.format(input_suffix))[0]
    with open(flagstat_file) as flag_handle:
        for line in flag_handle:
            if "mapped" in line:
                result_list.append(line.rstrip())
            elif "properly paired" in line:
                result_list.append(line.rstrip())
                break
    
    result_line = "|" + "|".join(result_list) + "|{0}\n".format(anchored_ref)
    summary_handle.write(result_line)
            
    summary_handle.close()

    logger.debug("finish summarize_results")

#
#   Clean up intermediate files
#

@follows(summarize_results)
@files(anchored_ref, anchored_ref_touchfile, bam_files)
def clean_intermediates(anchored_ref, anchored_ref_touchfile, bam_files):
    """ Clean up intermediate alignment files """
    
    logger.debug("start clean_intermediates")
    
    # Note - it might be nicer to just put all the intermediate files for 
    # a given round in a single sub-directory - then we could just delete that at the end
        
    if not opt.dirty:
        
        # script files, sai files, region list files
        intermediate_patterns = ['bwa_aln_*', 'bwa_sampe_*', 'samtools_mpileup_*',
                                 '*sai*', '*part.bcf*', '*region_*.txt', 'picard_*',
                                 '*.complete']
        
        for pattern in intermediate_patterns:
            for i in glob(pattern):
                os.unlink(i)
    
        # bwa index files
        for suffix in ['.pac', '.ann', '.amb', '.rpac', '.bwt', '.rbwt', '.sa', '.rsa']:
        
            index_file = cleaned_ref + suffix
            if os.path.exists(index_file):
                os.unlink(index_file)
    
        # Query-sorted BAM files
        for bam_file in bam_files:
            os.unlink(bam_file)
    
        # Anchor scaffold assemblies
        if os.path.exists('scaffold_assemblies'):
            shutil.rmtree('scaffold_assemblies')
    
    # Touch the completion file
    fh = open(anchored_ref_touchfile, 'w')
    
    logger.debug("finish clean_intermediates")
    return

@follows(picard_fixmate_wait)
@files(cleaned_ref, final_ref_raw, original_ref, sorted_bam_files, final_merged_bam)
def finalize_assembly(cleaned_ref, final_ref_raw, original_ref, sorted_bam_files, final_merged_bam):
    """ Restore fasta headers and produce final output file """
    
    logger.debug("start finalize_assembly")

    # Restore original headers to contigs, write single-line sequences
    # Note: this is mainly for downstream analysis by trans-ABySS
    args1 = [PYTHON_EXECUTABLE, REHEADER_SCRIPT, original_ref, cleaned_ref]
    logger.info(" ".join(args1))
    
    proc1 = subprocess.Popen(args1, shell=False, cwd=os.getcwd(),  
                            stderr=std_error_log, stdout=open(final_ref_raw, 'w'))
    proc1.wait()
    
    # Merge the final BAM files if there's more than one, otherwise just rename
    if len(sorted_bam_files) > 1:
        args2 = ['samtools', 'merge', final_merged_bam] + sorted_bam_files
        logger.info(" ".join(args2))
        proc2 = subprocess.Popen(args2, shell=False, cwd=os.getcwd(), stderr=std_error_log)
        proc2.wait()
    else:
        logger.info("mv {0} {1}".format(sorted_bam_files[0], final_merged_bam))
        os.rename(sorted_bam_files[0], final_merged_bam)

    # Index the final BAM file
    # Need to switch back to Picard here as well
    # proc3 = subprocess.Popen(['samtools', 'index', final_merged_bam], 
    #             shell=False, cwd=os.getcwd(), stderr=std_error_log)
    # proc3.wait()
    
    picard_cmd = """java -jar -Xmx10G -XX:MaxPermSize=10g {0}/BuildBamIndex.jar INPUT={1} VALIDATION_STRINGENCY=SILENT""".format(
                  PICARD_DIR, final_merged_bam)
    picard_args = picard_cmd.split(" ")
    logger.info(picard_cmd)
    
    proc3 = subprocess.Popen(picard_args, shell=False, cwd=os.getcwd(), stderr=std_error_log)
    proc3.wait()

    # Clean up a few last files
    clean_intermediates(anchored_ref, anchored_ref_touchfile, bam_files)

    logger.debug("finish finalize_assembly")
    return
    
def main():
    """ Main function to specify Ruffus pipeline target """
    
    if not opt.finalize:
    
        pipeline_printout(sys.stdout, [clean_intermediates], verbose = 1)
        # pipeline_printout_graph(
        #     open('polishing_pipeline_graph.svg','w'), "svg", [finalize_assembly])
        if not opt.recon:
            pipeline_run([clean_intermediates], verbose = 1, multiprocess = 1)

    else:
        pipeline_printout(sys.stdout, [finalize_assembly], verbose = 1)
        # pipeline_printout_graph(
        #     open('polishing_pipeline_graph.svg','w'), "svg", [finalize_assembly])
        if not opt.recon:
            pipeline_run([finalize_assembly], verbose = 1, multiprocess = 1)
        

if __name__ == '__main__':
    main()

