#!/usr/bin/env python
# encoding: utf-8
"""
scaffold_filler.py

Created by Rod Docking on 2010-06-30.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import subprocess
from optparse import OptionParser

from lib.common_utils import build_seq_dict, reverse_complement, read_single_entry_fasta, long_seq_to_lines_of_fifty
from lib.cluster_job import Cluster_job
from lib.blat_parsing import cat_blat, read_blat_file, check_blat_hit, estimate_gap_or_overlap_distance

PYTHON_EXECUTABLE  = sys.executable
ORIENT_SCRIPT      = os.path.dirname(sys.argv[0]) + '/lib/sam_to_oriented_fasta.py'
TRIMMER_SCRIPT     = os.path.dirname(sys.argv[0]) + '/lib/fasta_trimmer.py'

def check_scaffold_assembly(gap_dict, key, gap, seq_path, new_contig_seq, subcontig_list, result_dict):
    """Check a scaffold assembly for completeness"""

    print "Length of new contig sequence: ", len(new_contig_seq)
    print "Remaining subcontigs: ", len(subcontig_list)
    
    # File names
    assembly_file_root = "{0}/{1}_{2}".format(seq_path, key, gap_dict[key].index(gap))
    front_fwd_blat_name = assembly_file_root + '_front_fwd.psl'
    back_fwd_blat_name = assembly_file_root + '_back_fwd.psl'
    front_rev_blat_name = assembly_file_root + '_front_rev.psl'
    back_rev_blat_name = assembly_file_root + '_back_rev.psl'    
    merge_blat_name = assembly_file_root + '_merge.psl'
    assembly_file_fwd = assembly_file_root + "_assembled_fwd.fa"
    assembly_file_rev = assembly_file_root + "_assembled_rev.fa"

    # Print out assembled contigs
    forward_assembled_contig = read_single_entry_fasta(assembly_file_fwd)
    print "Forward assembly: {0}bp ending with: {1}".format(len(forward_assembled_contig), forward_assembled_contig[-50:])
    
    reverse_assembled_contig = read_single_entry_fasta(assembly_file_rev)
    reverse_assembled_contig_rc = reverse_complement(reverse_assembled_contig)
    print "Reverse assembly: {0}bp starting with: {1}".format(len(reverse_assembled_contig_rc), reverse_assembled_contig_rc[:50])
    
    #########################################################################
    # CASE 1 - Forward assembly cleanly matches front and back targets      #
    #########################################################################
    
    # Check the forward assembly against the front target
    print "\n==> Case 1: Forward assembly cleanly matches front and back targets <=="
    print "\n\tForward assembly to front target:"             
    hit_list = read_blat_file(front_fwd_blat_name, verbosity = 'Full')
    hit = hit_list[0] if len(hit_list) > 0 else ''
                
    if (not hit == '') and (hit['strand'] == '+') and (hit['t_end'] == hit['t_size']) and \
        (hit['num_mismatch'] <= 3) and (hit['blockcount'] == 1) and (hit['q_start'] == 0) and \
        ((hit['q_size'] - hit['q_end']) > 1):
        
        start = hit['q_end']
        front_extension_seq = forward_assembled_contig[start:]
        print "\tExtension by {0}bp,".format(hit['q_size'] - hit['q_end']),
        print " starting from base {0} on the query sequence".format(start)
        print "\t", front_extension_seq
        forward_extends_front = True                
    else:
        print "\t==> No extension"
        front_extension_seq = ''
        forward_extends_front = False 

    # Forward assembly vs. back target
    print "\n\tForward assembly to back target:"             
    hit_list = read_blat_file(back_fwd_blat_name, verbosity = 'Full')            
    hit = hit_list[0] if len(hit_list) > 0 else ''
    
    if (not hit == '') and (hit['strand'] == '+') and (hit['q_end'] == hit['q_size']) and \
        (hit['num_mismatch'] <= 3) and (hit['blockcount'] == 1) and (hit['t_start'] == 0):
        
        print "\t==> Forward extends back"
        stop = hit['q_start']
        forward_extends_back = True                
    else:
        print "\t==> No extension"
        forward_extends_back = False
    
    # If the forward assembly matches both targets, return the gap sequence
    if forward_extends_front and forward_extends_back:

        scaffold_seq = forward_assembled_contig[start:stop]
        print "\tScaffold sequence is from", start, " to ", stop
        if scaffold_seq == '':
            print "\tStrange alignment coordinates - continuing"
        else:    
            print "\n\tForward assembly closes the gap!"
            print "\t", scaffold_seq
            new_contig_seq = new_contig_seq + scaffold_seq + subcontig_list.pop()
            result_dict['num_merge'] += 1
            sys.stderr.write('|  Merge - forward assembly\n')
            return new_contig_seq, result_dict

    #########################################################################
    # CASE 2 - Reverse assembly cleanly matches front and back targets      #
    #########################################################################

    # Check the reverse assembly against the back target
    print "\n==> Case 2: Reverse assembly cleanly matches front and back targets <=="
    print "\n\tReverse assembly to back target:"                    
    hit_list = read_blat_file(back_rev_blat_name, verbosity = 'Full')
    hit = hit_list[0] if len(hit_list) > 0 else ''

    if (not hit == '') and (hit['strand'] == '-') and (hit['t_start'] == 0) and \
        (hit['num_mismatch'] <= 3) and (hit['blockcount'] == 1) and (hit['q_start'] == 0) and \
        ((hit['q_size'] - hit['q_end']) > 1):

        start = hit['q_end']
        back_target_end = hit['t_end']
        print "\tExtension by {0}bp, ".format(hit['q_size'] - hit['q_end']),
        print " starting from base {0} on the query sequence".format(start)
        back_extension_seq = reverse_assembled_contig[start:]
        back_extension_seq = reverse_complement(back_extension_seq)
        print "\t", back_extension_seq
        reverse_extends_back = True
    else:
        print "\t==> No extension"
        back_extension_seq = ''
        reverse_extends_back = False

    print "\n\tReverse assembly to front target:"             
    hit_list = read_blat_file(front_rev_blat_name, verbosity = 'Full')
    hit = hit_list[0] if len(hit_list) > 0 else ''
    
    if (not hit == '') and (hit['strand'] == '-') and (hit['q_end'] == hit['q_size']) and \
        (hit['num_mismatch'] <= 3) and (hit['blockcount'] == 1) and (hit['t_end'] == hit['t_size']):
        
        print "\t==> Reverse extends front"
        stop = hit['q_start']
        reverse_extends_front = True                
    else:
        print "\t==> No extension"
        reverse_extends_front = False

    # If the reverse assembly matches both targets, return the gap sequence
    if reverse_extends_back and reverse_extends_front:

        print "\tScaffold sequence is from", start, " to ", stop
        scaffold_seq = reverse_complement(reverse_assembled_contig[start:stop])
        
        if scaffold_seq == '':
            print "\tStrange alignment coordinates - continuing"
        else:    
            print "\n\tReverse assembly closes the gap!"
            print "\t", scaffold_seq
            new_contig_seq = new_contig_seq + scaffold_seq + subcontig_list.pop()
            result_dict['num_merge'] += 1
            sys.stderr.write('|  Merge - reverse assembly\n')
            return new_contig_seq, result_dict

    #########################################################################
    # CASE 3 - Direct overlap between forward and reverse assemblies        #
    #########################################################################

    # Check for an exact overlap between the two assemblies (that BLAT might have missed)
    print "\n==> Case 3: Direct overlap between forward and reverse assemblies, both assemblies extend into gap <=="
    simple_overlap_distance = simple_overlap_check(forward_assembled_contig, reverse_assembled_contig_rc)
    if simple_overlap_distance:
        print "\tAppears to be a simple overlap of {0} bases".format(simple_overlap_distance) 

    if simple_overlap_distance and front_extension_seq and back_extension_seq:
        print "\tClean merge:"
        
        # Overlap distance is the length that the two sequences overlap by
        print "\tExtend {0} bp from front flank".format(len(front_extension_seq))
        print "\tOverlap distance: ", simple_overlap_distance
        print "\tThen continue for another ", len(reverse_assembled_contig_rc[simple_overlap_distance:]), " on the back assembled sequence"
        print "\tUntil rejoining the back target at ", back_target_end
        
        # Scaffold sequence is what we need: 
        scaffold_seq = front_extension_seq + reverse_assembled_contig_rc[simple_overlap_distance:]
        print "\t", scaffold_seq
        print "\tNew contig length:", len(new_contig_seq)
        new_contig_seq = new_contig_seq + scaffold_seq + subcontig_list.pop()[back_target_end:]
        print "\tNew contig length: ", len(new_contig_seq)
        result_dict['num_merge'] += 1
        sys.stderr.write('|  Merge - direct overlap\n')
        return new_contig_seq, result_dict
    
    else:
        print "\tStrange alignment coordinates: continuing"
    
    #########################################################################
    # CASE 4 - Alignment overlap between forward and reverse assemblies     #
    #########################################################################
    
    print "\n==> Case 4: Alignment overlap between forward and reverse assemblies <=="
    # Check whether we can actually close the gap
    overlap_success, overlap_distance, scaffold_coords = check_merge_blat(
        merge_blat_name, front_fwd_blat_name, back_fwd_blat_name, front_rev_blat_name, back_rev_blat_name)
                    
    # Write the new extension sequence
    if overlap_success and (front_extension_seq or back_extension_seq):
        
        print "\tMerging contigs"
        
        # If we have a +ve distance overlap, insert the scaffold sequence...
        if overlap_distance > 0:
            scaffold_seq = forward_assembled_contig[scaffold_coords[0]:scaffold_coords[1]]
            print scaffold_seq, len(scaffold_seq), overlap_distance
            new_contig_seq = new_contig_seq + scaffold_seq + subcontig_list.pop()
        
        # Otherwise trim the trailing contig
        else:
            new_contig_seq = new_contig_seq + subcontig_list.pop()[(overlap_distance*-1):]
        
        sys.stderr.write('|  Merge - blat alignments overlap\n')
        result_dict['num_merge'] += 1
        return new_contig_seq, result_dict

    else:
        print "\tStrange alignment coordinates: continuing"

    #########################################################################
    # CASE 5 - No merge is possible, attempt to extend contig flanks        #
    #########################################################################
    
    print "\n==> Case 5: Extending contig flanks <=="
    original_gap_size = gap[1]-gap[0]+1
    print "\n\tFront extension: {0} {1}bp".format(front_extension_seq, len(front_extension_seq))
    print "\tOriginal gap was {0}bp".format(original_gap_size)
    print "\tBack extension: {0} {1}bp".format(back_extension_seq, len(back_extension_seq))
    print "\tPad with {0} N's".format(original_gap_size - len(front_extension_seq) - len(back_extension_seq))
    pad_size = original_gap_size - len(front_extension_seq) - len(back_extension_seq)
    
    # Extension from the front or back
    if (pad_size > 0) and (  (len(front_extension_seq) > 0) or (len(back_extension_seq) > 0) ):
        pad_sequence = ('N'*pad_size)
        scaffold_seq = front_extension_seq + pad_sequence + back_extension_seq
        sys.stderr.write('|  Extension\n')
        result_dict['num_extend'] += 1

    # No extension
    elif pad_size > 0:
        pad_sequence = ('N'*pad_size)
        scaffold_seq = pad_sequence
        sys.stderr.write('|  Fail - under-assembly\n')
        result_dict['num_fail'] += 1

    # For cases where the ABySS gap estimate may be a slight underestimate, insert both extensions plus 5 N's
    elif pad_size > -25:
        pad_sequence = 'N'*5
        scaffold_seq = front_extension_seq + pad_sequence + back_extension_seq
        sys.stderr.write('|  Extension - gap underestimate\n')
        result_dict['num_extend'] += 1

    # If the extensions are much longer than the estimated gap size, return the original gap
    else:
        print "\tPotential over-assembly!!!!"
        scaffold_seq = 'N'*original_gap_size
        sys.stderr.write('|  Fail - over-assembly\n')
        result_dict['num_fail'] += 1
    
    print "\t", scaffold_seq
    new_contig_seq = new_contig_seq + scaffold_seq + subcontig_list.pop()

    return new_contig_seq, result_dict


def check_merge_blat(merge_blat_name, front_fwd_blat_name, back_fwd_blat_name, front_rev_blat_name, back_rev_blat_name):
    """Check a blat file of the two assembled contigs to see if they overlap"""
    
    # Return early if the two assemblies don't overlap each other
    if (os.path.getsize(merge_blat_name) == 0):
        print "\n\t==> Assemblies don't overlap - not attempting to merge"
        return False, 0, [0, 0]
        
    # Inspect the relevant blat files
    back_fwd_hit_list = read_blat_file(back_fwd_blat_name)
    front_rev_hit_list = read_blat_file(front_rev_blat_name)
    print "\n\tReverse assembly to forward assembly:"
    merge_hit_list = read_blat_file(merge_blat_name, verbosity = 'Full')

    # Grab the other two hits as well
    front_fwd_list = read_blat_file(front_fwd_blat_name)
    if len(front_fwd_list) >= 1:
        front_fwd_hit = front_fwd_list[0]
    else:
        return False, 0, [0, 0]    
    
    back_rev_list = read_blat_file(back_rev_blat_name)
    if len(back_rev_list) >= 1:
        back_rev_hit = back_rev_list[0]
    else:
        return False, 0, [0, 0]    

    # ... or if there is more than one hit in any of the files
    if (len(merge_hit_list) > 1) or (len(back_fwd_hit_list) > 1) or (len(front_rev_hit_list) > 1):
        return False, 0, [0, 0]

    merge_hit = merge_hit_list[0]
        
    """
    Case 1:
                ------------------->
    XXXXXXXXXXXXXXXX             XXXXXXXXXXXXXXXXXXX
                <-------------------
    
    
    - fwd assembly overlaps reverse assembly
    - fwd assembly overlaps back target
    - rev assembly overlaps front target
    
    - scaffold seq is from where the front assembly leaves the front target, to where it enters the back target    
    """
    
    """
    Case 2:
                ------------------->
    XXXXXXXXXXXXXXXX                        XXXXXXXXXXXXXXXXXXX
                            <----------------------

    - fwd assembly overlaps reverse assembly
    - scaffold seq is from where the front assembly leaves the front target + the overlap section, plus where the 
        reverse assembly enters the back target
    """

    if (len(back_fwd_hit_list) == 1):

        back_fwd_hit = back_fwd_hit_list[0]
 
        # Check the two assemblies against each-other
        if ((merge_hit['num_mismatch'] <= 3) and (merge_hit['blockcount'] == 1) and \
            (merge_hit['t_end'] == merge_hit['t_size']) and (merge_hit['q_start'] == 0) and \
             (merge_hit['strand'] == '-')):
            print "\t==> Consistent overlap between the two assemblies"
    
            # Check the forward assembly vs. front and back targets
            overlap_list = [back_fwd_hit, front_fwd_hit]
            distance = estimate_gap_or_overlap_distance(overlap_list)
            print "\tDistance: ", distance
    
            if distance > 0:
                scaffold_coords = [front_fwd_hit['q_end'], back_fwd_hit['q_start']]            
            else:
                scaffold_coords = [0, 0]
        
            return True, distance, scaffold_coords     
        
        else:
            print "\n\tPoor overlap between assemblies"
            return False, 0, [0, 0]
        
    print "\tFound a potential merge!!!"
        
    return False, 0, [0, 0]

def extract_original_contig(reference, key, seq_path):
    """Extract the original contig sequence using samtools faidx"""
        
    process = subprocess.Popen(
        [ 'samtools', 'faidx', reference, key ],
        shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
    handle = process.stdout
    contig_handle = open("{0}/{1}_original.fa".format(seq_path, key),'w')
    for line in handle:
        contig_handle.write(line)
    handle.close()
    contig_handle.close()    
    
    return

def find_scaffold_locations(seq_dict, edits_list):
    """Parse the sequence dictionary, and find scaffolded gaps"""
    
    gap_dict = {}
    
    for key in seq_dict.iterkeys():

        seq = seq_dict[key]
        gap_dict[key] = []
        in_gap = False
        current_gap = [0, 0]
        
        for i in range(len(seq_dict[key])):
            
            if seq[i] == 'N' and (not in_gap):
                in_gap = True
                current_gap[0] = i
            elif seq[i] == 'N' and (in_gap):
                continue
            elif seq[i] != 'N' and (in_gap):
                in_gap = False
                current_gap[1] = i-1
                gap_dict[key].append(current_gap)
                current_gap = [0, 0]        
    
    num_gaps, num_contigs_with_gap, num_bases_in_gaps = 0, 0, 0
    keys_to_delete = []
    
    edit_handle = open(edits_list, 'a')
    
    # Go through the gap dictionary, and count gaps and bases in gaps
    for gap in gap_dict:
        if gap_dict[gap] == []:
            keys_to_delete.append(gap)
            continue

        # Write this contig to the edit list
        edit_handle.write("{0}\n".format(gap))
        
        num_contigs_with_gap += 1
        for indiv_gap in gap_dict[gap]:
            num_gaps += 1
            num_bases_in_gaps += (indiv_gap[1]-indiv_gap[0]+1)
    
    edit_handle.close()
    
    # Output header
    print num_gaps, " distinct gaps in ", num_contigs_with_gap, " distinct contigs, with ", num_bases_in_gaps, " bases in gaps"        
    sys.stderr.write("||{0:10}||{1:11}||{2:5}||{3:5}||{4:5}||{5:5}|| Assembly size front || Assembly size reverse\n".format(
        "Contig", "Gap Coord", "Est.", "R1", "R2", "R3"))

    print "ANCHOR_1:\t{0}\t{1}".format(num_gaps, num_bases_in_gaps)

    # Remove all contigs that don't have scaffolds for now
    for key in keys_to_delete:
        del gap_dict[key]
    
    return gap_dict, num_gaps            
    
def submit_scaffold_assembly(gap_dict, gap, key, seq, seq_path, seq_dict, bam_list, options):
    """Construct input files and submit an assembly job to the cluster"""
    
    assembly_file_root = "{0}/{1}_{2}".format(seq_path, key, gap_dict[key].index(gap))
    
    # Intermediate files
    read_file_fwd_tmp = assembly_file_root + "_fwd.fa.tmp"
    read_file_rev_tmp = assembly_file_root + "_rev.fa.tmp"
    read_file_fwd = assembly_file_root + "_fwd.fa"
    read_file_rev = assembly_file_root + "_rev.fa"
    assembly_file_fwd = assembly_file_root + "_assembled_fwd.fa"
    assembly_file_rev = assembly_file_root + "_assembled_rev.fa"
    front_flank_name = assembly_file_root + '_front_target.fa'
    back_flank_name = assembly_file_root + '_back_target.fa'
    front_fwd_blat_name = assembly_file_root + '_front_fwd.psl'
    back_fwd_blat_name = assembly_file_root + '_back_fwd.psl'
    front_rev_blat_name = assembly_file_root + '_front_rev.psl'
    back_rev_blat_name = assembly_file_root + '_back_rev.psl'    
    merge_blat_name = assembly_file_root + '_merge.psl'
    log_file = assembly_file_root + "_log.txt"
    
    log_handle = open(log_file, 'w')
    log_handle.write("Started {0}_{1}\n".format(key, gap_dict[key].index(gap)))
        
    #print key, gap
    big_window = "{0}:{1}-{2}".format(
        key, max(0,int(gap[0]-1000)), min(len(seq_dict[key]),int(gap[1]+1000)))
    
    small_window = "{0}-{1}".format(
        max(0,int(gap[0]-200)), min(len(seq_dict[key]),int(gap[1]+200)))

    # For each BAM file in the input list, extract reads:
    bam_files = []
    with open(bam_list, 'r') as bam_handle:
        for line in bam_handle:
            bam_files.append(line.rstrip())
    
    for bam_file in bam_files:
    
        # Extract reads from the first BAM file...
        # e.g. samtools view SRR001665_r0.bam 1982:49000-52000 | ~/svn/python_tools/sam_to_oriented_fasta_new.py -i215 49550-49750 > test.fa
        #print "samtools view {0} {1}".format(bam_list, big_window)    
        process_one = subprocess.Popen(
             [ 'samtools', 'view', bam_file, big_window ],
             shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
        log_handle.write("samtools view {0} {1} | \n".format(bam_file, big_window))    
    
        #print "/home/rdocking/seqdev/ALCAP/lib/sam_to_oriented_fasta_new.py -i400 -q3 {0} {1} {2}".format(
        #    small_window, read_file_fwd, read_file_rev)
        process_two = subprocess.Popen(
            [PYTHON_EXECUTABLE, ORIENT_SCRIPT, '-i{0}'.format(options.i_size), '-q{0}'.format(options.qtrim), small_window,
            read_file_fwd_tmp, read_file_rev_tmp],
            shell=False, cwd=os.getcwd(), stdin=process_one.stdout, stdout=subprocess.PIPE)
        process_two.wait()
        log_handle.write("{0} {1} -i{2} -q{3} {4} {5} {6}\n".format(PYTHON_EXECUTABLE, ORIENT_SCRIPT,
            options.i_size, options.qtrim, small_window, read_file_fwd_tmp, read_file_rev_tmp))

    # Trim reads
    trim_process = subprocess.call([PYTHON_EXECUTABLE, TRIMMER_SCRIPT, '-t{0}'.format(options.r_length), read_file_fwd_tmp, read_file_fwd])
    trim_process = subprocess.call([PYTHON_EXECUTABLE, TRIMMER_SCRIPT, '-t{0}'.format(options.r_length), read_file_rev_tmp, read_file_rev])
    log_handle.write("{0} {1} -t{2} {3} {4}\n".format(PYTHON_EXECUTABLE, TRIMMER_SCRIPT,
        options.r_length, read_file_fwd_tmp, read_file_fwd))
    log_handle.write("{0} {1} -t{2} {3} {4}\n".format(PYTHON_EXECUTABLE, TRIMMER_SCRIPT,
        options.r_length, read_file_rev_tmp, read_file_rev))
                               
    # Then assemble with DAssembler
    sys.stderr.write("|\t")
    process_three = subprocess.Popen(
        ['DAssembler', '-r{0}'.format(options.r_length), '-o{0}'.format(options.overlap), '-c{0}'.format(options.min_cov), read_file_fwd], 
        shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
    assembly_handle = open(assembly_file_fwd, 'a')
    assembly_handle.write(">forward_assembly\n")
    for line in process_three.stdout:
        assembly_handle.write(line)    
    assembly_handle.close()
    log_handle.write("DAssembler -r{0} -o{1} -c{2} {3}\n".format(options.r_length, options.overlap, options.min_cov, read_file_fwd))
    
    # Assemble the reverse direction
    sys.stderr.write("|\t")
    process_five = subprocess.Popen(
        ['DAssembler', '-r{0}'.format(options.r_length), '-o{0}'.format(options.overlap), '-c{0}'.format(options.min_cov), read_file_rev], 
        shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
    assembly_handle = open(assembly_file_rev, 'a')
    assembly_handle.write(">reverse_assembly\n")
    for line in process_five.stdout:
        assembly_handle.write(line)    
    assembly_handle.close()
    log_handle.write("DAssembler -r{0} -o{1} -c{2} {3}\n".format(options.r_length, options.overlap, options.min_cov, read_file_rev))
        
    # Each gap gets it's own cluster job
    #jobname = "scaffold_{0}_{1}.sh".format(key, gap_dict[key].index(gap))
    #qsub_header(name = jobname, directory = seq_path)

    # Find the inner and outer windows for retrieving reads
    outer_window = [int(i) for i in big_window.split(':')[1].split('-')]
                                
    # Find the flank coordinates and extract fasta files
    front_flank = [ outer_window[0], gap[0] ]
    back_flank = [ gap[1] + 1, outer_window[1] + 1]
                
    # Extract the flank sequence and write a temp file
    front_target = open(front_flank_name, 'w')
    target_seq = ">{0}_{1}_target_front\n{2}\n".format(key, gap_dict[key].index(gap), seq[front_flank[0]:front_flank[1]])
    front_target.write(target_seq)
    front_target.close()

    back_target = open(back_flank_name, 'w')
    target_seq = ">{0}_{1}_target_back\n{2}\n".format(key, gap_dict[key].index(gap), seq[back_flank[0]:back_flank[1]])
    back_target.write(target_seq)
    back_target.close()
    
    # BLAT the two assembled contigs against the flanks, and against each other
    #blat_command = "blat -noHead {0} {1} {2} >/dev/null".format(front_flank_name, assembly_file_fwd, front_fwd_blat_name)
    blat_process = subprocess.Popen(['blat', '-noHead', front_flank_name, assembly_file_fwd, front_fwd_blat_name],
                        shell=False, cwd=os.getcwd(), stdout=open('/dev/null', 'w'))
    blat_process.wait()                    
    #qsub_append(command = blat_command, name = jobname, directory = seq_path)

    #blat_command = "blat -noHead {0} {1} {2} >/dev/null".format(back_flank_name, assembly_file_fwd, back_fwd_blat_name)
    blat_process = subprocess.Popen(['blat', '-noHead', back_flank_name, assembly_file_fwd, back_fwd_blat_name],
                        shell=False, cwd=os.getcwd(), stdout=open('/dev/null', 'w'))
    blat_process.wait()                    

    #qsub_append(command = blat_command, name = jobname, directory = seq_path)

    #blat_command = "blat -noHead {0} {1} {2} >/dev/null".format(front_flank_name, assembly_file_rev, front_rev_blat_name)
    blat_process = subprocess.Popen(['blat', '-noHead', front_flank_name, assembly_file_rev, front_rev_blat_name],
                        shell=False, cwd=os.getcwd(), stdout=open('/dev/null', 'w'))
    blat_process.wait()                    

    #qsub_append(command = blat_command, name = jobname, directory = seq_path)
    
    #blat_command = "blat -noHead {0} {1} {2} >/dev/null".format(back_flank_name, assembly_file_rev, back_rev_blat_name)
    blat_process = subprocess.Popen(['blat', '-noHead', back_flank_name, assembly_file_rev, back_rev_blat_name],
                        shell=False, cwd=os.getcwd(), stdout=open('/dev/null', 'w'))
    blat_process.wait()                    

    #qsub_append(command = blat_command, name = jobname, directory = seq_path)
    #blat_command = "blat -noHead {0} {1} {2} >/dev/null".format(assembly_file_fwd, assembly_file_rev, merge_blat_name)
    blat_process = subprocess.Popen(['blat', '-noHead', assembly_file_fwd, assembly_file_rev, merge_blat_name],
                        shell=False, cwd=os.getcwd(), stdout=open('/dev/null', 'w'))
    blat_process.wait()                    

    #qsub_append(command = blat_command, name = jobname, directory = seq_path)
        
    # Submit the job for this contig - these are small enough to do locally at the moment
    #qsub_submit(name=jobname, local=True, directory = seq_path)

    log_handle.write("Finished {0}_{1}\n".format(key, gap_dict[key].index(gap)))
    log_handle.close()

    return  
    
def simple_overlap_check(seq1, seq2):
    """Check for a simple overlap between the end of seq1 and the front of seq2"""
    
    success = False
    
    # Use the last eight bases of the forward assembly as a seed
    seed_string = seq1[-8:]
    print "\n\tSeed: ", seed_string
    
    # Disallow seeds that consist of a single letter - allowing them leads to mis-assemblies
    if seed_string in ['A'*8, 'T'*8, 'C'*8, 'G'*8]:
        print "\tLow-complexity seed - returning"
        return success
    
    # Find the seed in the second sequence
    offset = seq2.rfind(seed_string)
    
    # If a match is found, check that the two strings agree before the match
    if (offset >= 0):
        print "\tFound seed from forwards assembly at", offset, " in reverse assembly"
        print "\t", seq1
        print "\t", " "* (len(seq1) - offset - 9), seq2
        
        if (seq1[-(offset+8):]) == (seq2[:(offset+8)]):
            print "\t==> DIRECT OVERLAP <=="
            success = len((seq2[:(offset+8)]))
    else:
        print "\tCouldn't find seed from forwards assembly in reverse assembly"
    
    return success        

def parse_options():
    """Parse options provided on the command line"""

    usage = "usage: %prog [options] reference.fa bam_files.txt output.fa edits_list.txt"
    parser = OptionParser(usage)

    parser.add_option("-q", "--qual_trim",
        action="store", type="int", dest="qtrim", 
        default=3, help="phred quality threshold for read trimming (assumes ASCII-33) [3]")
    parser.add_option("-o", "--overlap",
        action="store", type="int", dest="overlap", 
        default=20, help="DAssembler overlap parameter [20]")
    parser.add_option("-c", "--min_cov",
        action="store", type="int", dest="min_cov", 
        default=2, help="DAssembler minimum coverage parameter [2]")
    parser.add_option("-i", "--i_size",
        action="store", type="int", dest="i_size", 
        default=250, help="Mean fragment size [250]")
    parser.add_option("-r", "--read_length",
        action="store", type="int", dest="r_length", 
        default=100, help="Read length (longer reads will be trimmed) [100]")

    (options, args) = parser.parse_args()
    if len(args) < 4:
        parser.error("incorrect number of arguments")
        
    # Replace arguments with full path names
    pathed_args = [os.path.abspath(arg) for arg in args]

    # Check that all files exist
    for arg in pathed_args[:2]:
        if not os.path.exists(arg):
            parser.error("{0} does not exist!".format(arg))
    
    return options, pathed_args
    
def main():
    """Find scaffolds and attempt to fill them!"""

    options, args = parse_options()
    reference, bam_list, output_file, edits_list = args[0], args[1], args[2], args[3]

    # Retrieve the sequences as a dictionary 
    #   - (note that this might be slow for a large dictionary!)
    seq_dict = build_seq_dict(reference)
    
    # For each sequence, check for scaffold gaps, and note contigs to edit
    gap_dict, num_gaps = find_scaffold_locations(seq_dict, edits_list)
                
    # Go through the gap dictionary, and find coordinates to assemble
    #   Outer window limit: how much flanking sequence from the target contig to extract for use as a BLAT target
    #   Inner window limit: how much flanking sequence to try to assemble
    
    # Set up output directory
    scaffold_path = os.getcwd() + '/scaffold_assemblies'
    if not os.path.exists(scaffold_path):
        os.mkdir(scaffold_path)

    contig_counter = 0
    result_dict = {
        'num_merge' : 0,
        'num_extend' : 0,
        'num_fail' : 0
    }
    
    # Submit a set of directional assemblies for each contig with gaps
    for key in gap_dict.iterkeys():
                
        seq_path = scaffold_path + '/{0}'.format(key)
        if not os.path.exists(seq_path):
            os.mkdir(seq_path)
                
        # Extract the original contig
        extract_original_contig(reference, key, seq_path)
        seq_path = scaffold_path + '/{0}'.format(key)
        contig = read_single_entry_fasta("{0}/{1}_original.fa".format(seq_path, key))

        # Split the contig into subcontigs
        splits = contig.split('N')
        subcontig_list = filter(lambda x: x != '', splits)
        subcontig_list.reverse()
       
        # Start piecing together the new contig
        new_contig_seq = subcontig_list.pop()

        print key, len(seq_dict[key]), '\n----------------'
        contig_counter += 1
        print contig_counter, " of ", len(gap_dict.keys())
                
        # For each gap, assemble                
        for gap in gap_dict[key]:
            sys.stderr.write("|{0:12}|{1:12}|{2:6}|".format(key, str(gap).lstrip('[').rstrip(']'), (gap[1]-gap[0]+1)))
            seq = seq_dict[key]
            submit_scaffold_assembly(gap_dict, gap, key, seq, seq_path, seq_dict, bam_list, options)
                        
            print "\n\n\n", key, "{0}_{1}\n----------------------".format(key, gap_dict[key].index(gap))
            
            # Inspect the assembly, and return the growing contig sequence
            new_contig_seq, result_dict = check_scaffold_assembly(gap_dict, key, gap, seq_path, new_contig_seq, subcontig_list, result_dict)
                                        
        # Update the sequence dictionary
        seq_dict[key] = new_contig_seq
        
    # Write the new reference    
    out_handle = open(output_file,'w')
    for key in sorted(seq_dict.keys()):
        line_to_write = ">{0}\n{1}".format(key, long_seq_to_lines_of_fifty(seq_dict[key]))
        out_handle.write(line_to_write)
    out_handle.close()        
    
    print "Contigs checked: ", contig_counter
    print result_dict
    print "ANCHOR_2:\t{0}\t{1}\t{2}".format(result_dict['num_merge'], result_dict['num_extend'], result_dict['num_fail'])
                
if __name__ == '__main__':
    main()
