#!/usr/bin/env python
# encoding: utf-8
"""
cluster_job.py

Written by Rod Docking (rdocking@bcgsc.ca)
"""

import sys
import os
import subprocess
from time import sleep
from optparse import OptionParser
import socket

def parse_options(default_job):
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] \'commands_to_run\'"
    parser = OptionParser(usage)
    
    parser.add_option("-e", "--host",
        action="store", dest="host", type="string",
        help="Cluster head node name [{0}]".format(default_job.host), 
        default=default_job.host)
    parser.add_option("-d", "--dir",
        action="store", dest="directory", type="string",
        help="Starting directory [{0}]".format(default_job.directory),
        default=default_job.directory)
    parser.add_option("-j", "--jobname",
        action="store", dest="jobname", type="string",
        help="Job name [{0}]".format(default_job.jobname),
        default=default_job.jobname)
    parser.add_option("-w", "--dependencies",
        action="store", dest="dependencies", type="string",
        help="Comma-separated list of dependencies [{0}]".format(default_job.dependencies),
        default=default_job.dependencies)
    parser.add_option("-q", "--queue",
        action="store", dest="queue", type="string",
        help="Job queue [{0}]".format(default_job.queue),
        default=default_job.queue)
    parser.add_option("-p", "--parallel",
        action="store", dest="pe", type="string",
        help="Parallel environment and cpus [{0}]".format(default_job.pe),
        default=default_job.pe)
    parser.add_option("-n", "--cpus",
        action="store", dest="cpus", type="string",
        help="Number of cpus to request [{0}]".format(default_job.cpus),
        default=default_job.cpus)
    parser.add_option("-m", "--mem_token",
        action="store", dest="mem_token", type="string",
        help="Memory reservation [{0}]".format(default_job.mem_token),
        default=default_job.mem_token)
    parser.add_option("-t", "--tmp_token",
        action="store", dest="tmp_token", type="string",
        help="Temp space reservation [{0}]".format(default_job.tmp_token),
        default=default_job.tmp_token)
    parser.add_option("-f", "--cmd_file",
        action="store", dest="cmd_file", type="string",
        help="File with commands to run", 
        default=None)
    parser.add_option("-B", "--batch",
        action="store_true", dest="batch",
        help="Each line in CMD_FILE should be run as a separate job", 
        default=None)
    # parser.add_option("-N", "--no_submit",
    #     action="store_true", dest="no_submit",
    #     help="Generate job submission scripts, but don't submit the jobs", 
    #     default=None)


    
    (options, args) = parser.parse_args()
    
    # if not (len(args) == 3):
    #     parser.error("incorrect number of arguments")
                
    return options, args

class Cluster_job:
    """A class for creating and submitting cluster jobs"""
    
    def __init__(self, host = None, jobname = (os.environ.get("USER") + '_job.sh'), directory = os.getcwd(),
                dependencies = None, queue = 'all.q', pe = None, cpus = None, 
                mem_token = "2G", tmp_token = None, no_submit = False):
        """Create the header code for the submission script"""
        
        # Try to figure out which host to submit on
        self.host = host
        self.localhost = socket.gethostname()
    
        # If a head-node host isn't specified, try to infer the correct one
        if not self.host:
            if 'xhost' in self.localhost:
                self.host = 'apollo'
            elif 'login' in self.localhost:
                self.host = 'login4'
            elif 'node' in self.localhost:
                self.host = 'login4'
            else:
                self.host = 'apollo'    
        
        # Get fully-pathed job name
        self.jobname = jobname
        self.directory = directory
        self.base_jobname = os.path.basename(jobname)
        self.pathed_jobname = os.path.join(directory, self.base_jobname)

        # If the job directory doesn't exist, create it
        if not os.path.exists(self.directory):
           os.makedirs(self.directory)

        self.dependencies = dependencies
        self.queue = queue
        self.pe = pe
        self.cpus = cpus
        self.mem_token = mem_token
        self.tmp_token = tmp_token
        self.no_submit = no_submit
        
        self.job_lines = []
        
    def append(self, cmd):
        """Append a command to the job"""
        
        self.job_lines.append("{0}".format(cmd))
    
    def submit(self):
        """Submit the job"""

        # Open the script file and add header info
        self.job_handle = open(self.pathed_jobname, 'w')
        self.job_handle.write('#! /bin/sh\n#$ -S /bin/sh\n#$ -N {0}\n#$ -V\n#$ -R y\n'.format(self.jobname))
        self.job_handle.write('#$ -wd {0}\n'.format(self.directory))

        # Split and re-join dependencies list to make sure it's comma-separated
        if self.dependencies:
            if len(self.dependencies.split()) > 1:
                self.dependencies = ",".join(self.dependencies)
            self.job_handle.write("#$ -hold_jid {0}\n".format(self.dependencies))

        # Set queue
        self.job_handle.write("#$ -q {0}\n".format(self.queue))
        
        # Set parallel environment options
        if self.pe and self.cpus:
            self.job_handle.write("#$ -pe {0} {1}\n".format(self.pe, self.cpus))
        
        # Set memory and tmp space reservations
        if self.tmp_token:
            self.job_handle.write("#$ -l tmp_token={0}\n".format(self.tmp_token))
        self.job_handle.write("#$ -l mem_token={0}\n\n\n".format(self.mem_token))
                
        # cd to cwd
        self.job_handle.write("cd {0}\n".format(self.directory))        
        
        # Add job commands
        self.job_handle.write("\n".join(self.job_lines))
        self.job_handle.write("\necho Job complete!\n")
        self.job_handle.close()
        
        # ssh to head node and submit job
        proc = subprocess.Popen(
            ['ssh', self.host, 'qsub', self.pathed_jobname],
            shell=False, cwd=os.getcwd(), stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            )
        proc.wait()

        # Grab output from cluster and store job id
        self.submission_lines_stdout =  proc.stdout.readlines()
        self.submission_lines_stderr = proc.stderr.readlines()

        print self.submission_lines_stdout
        print self.submission_lines_stderr
    
        # Get job id:
        if "Your job" in self.submission_lines_stdout[0]:
            self.job_id = int(self.submission_lines_stdout[0].split()[2])
        else:
            self.job_id = None    

        sleep(2)
        

if __name__ == "__main__":
    """ Self-test code for the Cluster_job class """

    # Create a dummy object to get the defaults (so we're not editing them in two places)
    #  If the user updates any options, those will be provided in place of the defaults
    default_job = Cluster_job()
    opt, args = parse_options(default_job)

    # Batch mode - create a job for each line in the command file (similar to mqsub)
    if opt.cmd_file and opt.batch:
        
        iterator = 0
        
        for line in open(opt.cmd_file, 'r'):            
            # Create the job to submit
            C = Cluster_job(host = opt.host, jobname = opt.jobname + '_' + str(iterator), 
                        directory = opt.directory, dependencies = opt.dependencies, 
                        queue = opt.queue, pe = opt.pe, cpus = opt.cpus, 
                        mem_token = opt.mem_token, tmp_token = opt.tmp_token)
            C.append(line.rstrip())
            
            #if not options.no_submit:
            C.submit()
            
            iterator += 1            
    
    # Regular mode - either submit a job with the command line arguments
    #  or with all the lines of the input file as a single job            
    else:

        # Create the job to submit
        C = Cluster_job(host = opt.host, jobname = opt.jobname, directory = opt.directory,
                    dependencies = opt.dependencies, queue = opt.queue, pe = opt.pe, cpus = opt.cpus, 
                    mem_token = opt.mem_token, tmp_token = opt.tmp_token)

        if opt.cmd_file:
            for line in open(opt.cmd_file, 'r'):
                C.append(line.rstrip())
        
        else:    
            C.append(args[0])
    
        # Submit the job
        #if not options.no_submit:
        C.submit()

    
    