#!/home/sequence/bioinformatics/seqdev/CentOS_54/python/Python-2.6.4/bin/python2.6
# encoding: utf-8
"""
fasta_laundry.py

Written by Rod Docking (rdocking@bcgsc.ca)
"""

import sys
import re
import string
from optparse import OptionParser

#Define exceptions
class FastaError(Exception): pass
class HeaderError(FastaError): pass
class SeqError(FastaError):
    def __init__(self, expr):
        self.expr = expr
        print "Something looks wrong with this sequence!\n{0}".format(expr)
        sys.exit()

fastaHeaderPattern = re.compile('^>\S{1,}')
seq_on_lineHeaderPattern = re.compile('^>\S{1,}.+[ATCGatcg]{5,}$')
ambiguousSeqPattern = re.compile('[^ATCGNX?\.atcgnx\n]')
ambiguousSeqPatternStrict = re.compile('[^ATCGatcg\n]')

# Helpers for the 'ambiguation' function
def sub_N(match_object): return 'N' 
ambiguous_bases = re.compile(r'[MRWSYKVHDBmrwsykvhdb]')

class FastaRecord(object):
    """Simple class to hold a single fasta record"""

    def __init__(self, raw_header, raw_seq_list):
        self.raw_header = raw_header
        self.header_id, self.header_comment = split_header(raw_header)

        self.raw_seq_list = raw_seq_list
        self.long_seq = "".join(raw_seq_list)
        
    def __str__(self):
        return "{0}\n{1}\n".format(self.raw_header, "".join(self.raw_seq_list))
        

def parse_options():
    """Parse options provided on the command line"""

    usage = """usage: %prog [options] fasta_in fasta_out
        
        fasta_in and fasta_out can be replaced with '-' and ' '
        for stdin and stdout"""
    parser = OptionParser(usage)
    parser.add_option("-v", "--verbose",
        action="store_true", dest="verbose", default=False, 
        help="print verbose output")
    parser.add_option("-t",
        action="store", type="int", dest="trim", 
        help="trim any sequences shorter than this threshold")
    parser.add_option("-u",
        action="store", type="int", dest="untrim", 
        help="trim any sequences LONGER than this threshold")
    parser.add_option("-i",
        action="store", type="string", dest="seq_iterator", 
        help="replace sequence ids with this string plus an iterator (use 'num' for just the counter)",)
    parser.add_option("-l",
        action="store", type="int", dest="line_length", 
        help="write sequences with a specified line length",
        default = 50)
    parser.add_option("-f",
        action="store", type="int", dest="flank", 
        help="prepend each sequence by a specified number of N's")
    parser.add_option("-a",
        action="store_true", dest="ambiguate", 
        help="Convert all non-ACGT characters to N's")
    parser.add_option("-R",
        action="store_true", dest="reverse_complement", 
        help="reverse-complement all entries, and reverse their order in the file")
    parser.add_option("-S",
        action="store_true", dest="single_line", 
        help="write each sequence on a single line")
    parser.add_option("-N",
        action="store_true", dest="no_comment", 
        help="strip all comment fields")
    parser.add_option("-T",
        action="store_true", dest="strict", 
        help="Allow only ATCG characters")
    parser.add_option("-x",
        action="store", type="string", dest="exclude_dict", 
        help="exclude any sequences with ids from this list",) 
    parser.add_option("-c",
        action="store", type="string", dest="include_list", 
        help="include only sequences with ids from this list",)               
    parser.add_option("-r",
        action="store", type="string", dest="rc_dict", 
        help="reverse-complement sequences with ids from this list",)               
        
    (options, args) = parser.parse_args()
    
    if(len(args) == 0):
        parser.error("incorrect number of arguments")
    
    # if options.single_line and options.line_length:
    #     parser.error("-S and -l are mutually exclusive")
    
    return options, args

def reverse_complement(seq):
    """Return the reverse complement"""

    return seq.translate(string.maketrans("ATCGatcgNnXx", "TAGCtagcNnXx"))[::-1]

def split_header(header_to_split):
    """Take a fasta header and return header name and comment"""
    
    if not header_to_split:
        raise HeaderError, 'Input cannot be blank'
    if not fastaHeaderPattern.search(header_to_split):
        raise HeaderError, 'Header {0} looks wrong'.format(header_to_split)
    if seq_on_lineHeaderPattern.search(header_to_split):
        raise HeaderError, 'Header {0} looks wrong'.format(header_to_split)
    
    raw_header = header_to_split.lstrip('>')
    header_tuple = raw_header.partition(" ")
    return header_tuple[0], header_tuple[2].rstrip()

def analyze_last_seq(F, counter, options, exclude_dict, include_list, rc_dict):
    """Analyze fasta sequence, returning filtered sequence in fasta format"""
    
    if not F.long_seq:
        raise SeqError(F.long_seq), 'Seq cannot be blank'
    if options.strict and ambiguousSeqPatternStrict.search(seq):
        raise SeqError(F.long_seq), 'Seq looks strange: {0}'.format(seq)
            
    # seq_lines = seq.split("\n")
    # long_seq, seq_to_return = '', ''
    # for line in seq_lines:
    #     long_seq = long_seq + line.rstrip()
        
    if options.verbose:
        sys.stderr.write("Header: {0} Comment: {1} Length: {2}".format(
            F.header_id, F.header_comment, len(F.long_seq))) 
    
    if options.exclude_dict and (F.header_id in exclude_dict):
        if options.verbose:
            sys.stderr.write(" ==> EXCLUDED")
        return False        
    
    if options.include_list and (F.header_id not in include_list):
        if options.verbose:
            sys.stderr.write(" ==> EXCLUDED")
        return False
    
    if options.rc_dict and (F.header_id in rc_dict):
        if options.verbose:
            sys.stderr.write(" ==> REVERSE-COMPLEMENT")
        F.long_seq = reverse_complement(F.long_seq)
        #seq = reverse_complement(seq)
    
    if options.trim:
        if len(F.long_seq) < options.trim:
            if options.verbose:
                sys.stderr.write(" ==> TRIMMED (too short!)\n") 
            return False
    
    if options.untrim:
        if len(F.long_seq) > options.untrim:
            if options.verbose:
                sys.stderr.write(" ==> TRIMMED (too long!)\n") 
            return False
    
    if options.ambiguate:
        F.long_seq = re.sub(ambiguous_bases, sub_N, F.long_seq)        
    
    if options.flank:
        F.long_seq = "N"*options.flank + F.long_seq 
        #seq = "N"*options.flank + seq
    
    if options.single_line:
        seq_to_return = F.long_seq
    else:

        num_lines = len(F.long_seq) / options.line_length
        last_line = len(F.long_seq) - (num_lines*options.line_length)
                
        split_list = []
        for i in range(num_lines):
            trim_line = F.long_seq[(i * options.line_length):(i * options.line_length) + options.line_length]
            split_list.append(trim_line)
        
        if not last_line == 0:
            trim_line = F.long_seq[-last_line:]
            split_list.append(trim_line)

        seq_to_return = "\n".join(split_list)

    if options.seq_iterator == 'num':
        F.header_id = "{0}".format(counter)
    elif options.seq_iterator:
        F.header_id = "{0}{1}".format(options.seq_iterator, counter)
    
    if options.verbose:   
        sys.stderr.write("\n")
    
    if (options.no_comment) or (F.header_comment == ''):
        return ">{0}\n{1}\n".format(F.header_id, seq_to_return)
        
    return ">{0} {1}\n{2}\n".format(F.header_id, F.header_comment, seq_to_return)

def write_rc_file(original_outfile):
    """Re-write the output file with all sequences in rc orientation, and in the reverse order"""
    
    rc_list = []
    current_header, current_seq, counter = '', '', 0
    
    in_handle = open(original_outfile, 'r')
    for line in in_handle.readlines():
        if (current_seq != '') and (">" in line):
            seq = current_seq.lstrip('\n').rstrip('\n')
            rc_list.append([current_header.rstrip(), reverse_complement(seq)])
            current_header, current_seq = '', ''
            
        if ">" in line:
            current_header = line
            counter += 1
        else:
            current_seq = current_seq + line        
            
    seq = current_seq.lstrip('\n').rstrip('\n')
    rc_list.append([current_header.rstrip(), reverse_complement(seq)])
    in_handle.close()
    
    rc_list.reverse()
    
    out_handle = open(original_outfile,'w')
    for entry in rc_list:
        out_handle.write("{0}\n{1}\n".format(entry[0], entry[1]))
    out_handle.close()    

def main():
    """Trim, clean, or re-name sequences in a fasta file"""
    
    # Parse command-line options
    options, args = parse_options()
    
    # Figure out whether to write to files or stdin/out    
    fasta_in, fasta_out = '', ''
    if args[0] == '-':
        fasta_in = sys.stdin
    else:
        fasta_in = open(args[0],'r')    
    
    if len(args) == 2:
        fasta_out = open(args[1],'w')
    else:
        fasta_out = sys.stdout

    # Gather lists of sequences to exclude, include, or reverse-complement
    exclude_dict = {}
    if options.exclude_dict:
        for line in open(options.exclude_dict,'r'):
            exclude_dict[str(line.rstrip())] = 1
        print "Found {0} sequences to exclude".format(len(exclude_dict))
        
    include_list = []    
    if options.include_list:
        for line in open(options.include_list,'r'):
            include_list.append(line.rstrip())
        print "Found {0} sequences to include".format(len(include_list))
    
    rc_dict = {}
    if options.rc_dict:
        for line in open(options.rc_dict,'r'):
            rc_dict[str(line.rstrip())] = 1
        print "Found {0} sequences to reverse-complement".format(len(rc_dict))
         
    current_header, current_seq_list, counter = '', [], 0    
    for line in fasta_in:
        
        if (current_seq_list != []) and (">" in line):
            
            F = FastaRecord(current_header, current_seq_list)
            seq_to_write = analyze_last_seq(F, counter, options, 
                            exclude_dict, include_list, rc_dict)
            if seq_to_write:
                fasta_out.write(seq_to_write)
            else:
                counter -= 1    
            current_header, current_seq_list = '', []
            
        if ">" in line:
            current_header = line.rstrip()
            counter += 1
        else:
            current_seq_list.append(line.rstrip())        

    F = FastaRecord(current_header, current_seq_list)
    seq_to_write = analyze_last_seq(F, counter, options, 
                        exclude_dict, include_list, rc_dict)
    if seq_to_write:
        fasta_out.write(seq_to_write)

    # Close file-handles
    fasta_in.close()
    fasta_out.close()
    
    # Write the reverse-complement file
    if options.reverse_complement:
        write_rc_file(args[1])

if __name__ == "__main__":
    
    main()
