#!/usr/bin/env python
# encoding: utf-8
"""
generate_test_set.py

Created by Rod Docking on 2011-03-25.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import random
import gzip
import subprocess
from optparse import OptionParser

from common_utils import long_seq_to_lines_of_fifty, reverse_complement

COVERAGE = 20
FRAGMENT_LENGTH = 500
READ_LENGTH = 100
SCAFFOLD_LENGTH = 200

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] ref.fa OUTPUT_ROOT"
    parser = OptionParser(usage)
    
    parser.add_option("-u",
        action="store", type="float", dest="point",
        help="Per-base rate of point mutations to induce [0.001]",
        default = 0.001)
    parser.add_option("-i",
        action="store", type="float", dest="indel",
        help="Per-base rate of indels to induce [0.0001]",
        default = 0.0001)
    parser.add_option("-s",
        action="store", type="int", dest="scaffold",
        help="Number of scaffolds to introduce [10]",
        default = 10)
    parser.add_option("-c",
        action="store", type="int", dest="contigs",
        help="Number of contigs to split to [10]",
        default = 10)
    
    (opt, args) = parser.parse_args()
    
    if not (len(args) == 2):
        parser.error("incorrect number of arguments")
    
    return opt, args

def generate_fragments(output_root, contig_seq):
    """  
    Generate fake reads and output as BAM and gzipped fastq
    """

    num_fragments = len(contig_seq) * COVERAGE * 2 / FRAGMENT_LENGTH
    sys.stderr.write("Generating {0} fragments for {1}x coverage\n".format(num_fragments, COVERAGE))
    
    # Set up processes for writing BAM and fastq
    bam_proc = subprocess.Popen(['samtools', 'view', '-bS', '-'], 
                shell=False, cwd=os.getcwd(), 
                stdout=open('{0}.bam'.format(output_root), 'wb'),
                stdin=subprocess.PIPE)
    
    fastq_one = gzip.open('{0}_1.fq.gz'.format(output_root), 'wb')
    fastq_two = gzip.open('{0}_2.fq.gz'.format(output_root), 'wb')
        
    # Write BAM header    
    bam_proc.stdin.write("@HD\tSO unsorted\n")
    bam_proc.stdin.write("@SQ\tSN: 1\tLN: 1\n")
    
    # Generate reads
    for i in range(num_fragments):
        frag_loc = random.randrange(len(contig_seq) - FRAGMENT_LENGTH)
        fwd_seq = contig_seq[frag_loc:frag_loc + READ_LENGTH]
        rev_seq = reverse_complement(
                contig_seq[frag_loc + FRAGMENT_LENGTH - READ_LENGTH : 
                            frag_loc + FRAGMENT_LENGTH])
        
        assert(len(fwd_seq) == len(rev_seq) == READ_LENGTH)
        
        # Write as BAM
        sam_one = "read_{0}/1\t68\t*\t0\t0\t*\t*\t0\t0\t{1}\t{2}\n".format(
                   i, fwd_seq, "I"*READ_LENGTH)
        sam_two = "read_{0}/2\t132\t*\t0\t0\t*\t*\t0\t0\t{1}\t{2}\n".format(
                   i, rev_seq, "I"*READ_LENGTH)                 
        bam_proc.stdin.write(sam_one)
        bam_proc.stdin.write(sam_two)
        
        # Write as fastq
        fastq_one.write("@read_{0}/1\n{1}\n+\n{2}\n".format(i, fwd_seq, "I"*READ_LENGTH))
        fastq_two.write("@read_{0}/2\n{1}\n+\n{2}\n".format(i, rev_seq, "I"*READ_LENGTH))
        
    fastq_one.close()
    fastq_two.close()
    
    # Write input files
    fq_input_file = open('{0}.fastq.in'.format(output_root), 'w')
    fq_input_file.write("{0}_1.fq.gz\n{0}_2.fq.gz\n".format(output_root))
    fq_input_file.close()
    
    bam_input_file = open('{0}.bam.in'.format(output_root), 'w')
    bam_input_file.write("{0}.bam".format(output_root))
    bam_input_file.close()
            
    return

def induce_point_mutation(seq, opt):
    """ 
    Induce a point mutation at a random base in the sequence 
    """

    bases = ['A','C','G','T']
    position = random.randrange(len(seq))
    current_base = new_base = seq[position]
    
    while current_base == new_base:
        new_base = random.choice(bases)
    
    # Output as pseudo-vcf - the original base should be the ALT, 
    # since that's what the reads will have
    sys.stderr.write("I\t{0}\t.\t{1}\t{2}\n".format(position, new_base, current_base))

    return seq[:position] + new_base + seq[position+1:]
    
def induce_indel(seq, opt):
    """ 
    Induce an indel at a random base in the sequence 
    """

    bases = ['A','C','G','T']
    position = random.randrange(len(seq))
    
    # Randomly choose insertion or deletion
    if random.random() >= 0.5:
        seq = seq[:position] + random.choice(bases) + seq[position:]    # insertion
        sys.stderr.write("I\t{0}\t.\t{1}\t{2}\tINDEL\t(insertion)\t{3}\n".format(
                        position, seq[position-1:position+2], seq[position-1:position+3], len(seq)))
    else:
        seq = seq[:position] + seq[position+1:]                         # deletion        
        sys.stderr.write("I\t{0}\t.\t{1}\t{2}\tINDEL\t(deletion)\t{3}\n".format(
                        position, seq[position-1:position+2], seq[position-1:position+1], len(seq)))

    return seq    

def induce_scaffold(seq, opt):
    """ 
    Induce a scaffold at a random base in the sequence 
    """
    
    # Prevent a scaffold at the very beginning or end of the sequence
    position = random.randrange(len(seq)-2000) + 1000
    sys.stderr.write("I\t{0}\tSCAFFOLD\t{1}\n".format(position, len(seq)))
    
    return seq[:position] + 'N'*SCAFFOLD_LENGTH + seq[position+SCAFFOLD_LENGTH:]

def main():
    """ 
    From a reference sequence, introduce point mutations, indels, and scaffolds 
    """

    # Parse options
    opt, args = parse_options()
    input_file, output_root = args

    # Read the original sequence - this will concatenate any multi-entry FASTA files
    contig_seq_list = []
    with open(input_file, 'r') as fh:
        for line in fh:
            if ">" in line:
                sys.stdout.write(line)
            else:    
                contig_seq_list.append(line.rstrip())
    
    contig_seq = "".join(contig_seq_list).upper()
    sys.stderr.write("Contig is {0}bp\n".format(len(contig_seq)))            
    
    # Sample reads from the unmodified contigs
    generate_fragments(output_root, contig_seq)
    
    # Calculate the number of point mutations and indels to introduce
    num_point_mutations = int(opt.point * len(contig_seq))
    num_indels = int(opt.indel * len(contig_seq))
    sys.stderr.write("Inducing {0} point mutations and {1} indels\n".format(
                        num_point_mutations, num_indels))
    
    # Introduce random point mutations
    for i in range(num_point_mutations):
        contig_seq = induce_point_mutation(contig_seq, opt)
    
    # Introduce random indels    
    for i in range(num_indels):
        contig_seq = induce_indel(contig_seq, opt)
    
    # Induce scaffolds
    for i in range(opt.scaffold):
        contig_seq = induce_scaffold(contig_seq, opt)
    
    # Generate break-points for new contigs
    breakpoints = [0] + [random.randint(0, len(contig_seq)) for i in range(opt.contigs-1)] + [len(contig_seq)]
    breakpoints.sort()
    
    # Write out final contigs
    contig_handle = open('{0}-contigs.fa'.format(output_root), 'w')
    for i in range(opt.contigs):
        print i, breakpoints[i], breakpoints[i+1]
        
        # Slice the contig and trim and leading or trailing N characters
        contig = contig_seq[breakpoints[i]:breakpoints[i+1]]
        contig = contig.strip("N")
        contig_handle.write(">{0} {1}\n{2}\n".format(i,len(contig), contig))
    

if __name__ == '__main__':
    main()

